	;; Intel hex to binary utility
	;;
	;; Reads Intel hex records from the serial port and writes them to
	;; memory.
	;; 
	;; Copyright Simon Rowe <srowe@mose.org.uk> 2015


	; KERNAL entry points
SETLFS	= $FFBA
SETNAM	= $FFBD
OPEN	= $FFC0
CLOSE	= $FFC3
CHKIN	= $FFC6
CHRIN	= $FFCF
CHROUT	= $FFD2
SAVE	= $FFD8
CLALL	= $FFE7

	; Working storage
START	= $45			; Address of first record
END	= $47			; Address of last byte +1
ADDR	= $49			; Address to write data to
RECLEN	= $4B			; Record length
POS	= $4C			; Current position in record
CSUM	= $4D			; Accumlated checksum

*	= $2000
	.WORD *
	* = *-2

	CLD
	LDA #0
	STA START
	STA START+1
	LDA #2			; Open serial port
	LDX #<SERCTRL
	LDY #>SERCTRL
	JSR SETNAM
	LDA #128		; Send LF after CR
	LDX #2			; RS-232
	LDY #0
	JSR SETLFS
	JSR OPEN
	CMP #$F0		; non-standard error handling, Cb always set
	BNE OFAILED
	LDX #128		; Set serial as input
	JSR CHKIN
	BCS CFAILED
LOOP	JSR PROCREC		; Process records
	BCC LOOP
EXIT	LDA #128
	JSR CLOSE
	JSR CLALL		; Close and reset input
	RTS

OFAILED	LDA #<OPFS
	LDY #>OPFS
	JSR PRTSTR
	JSR CLALL
	RTS
CFAILED	LDA #<CKFS
	LDY #>CKFS
	JSR PRTSTR
	JMP EXIT


	; Process one record.
	;
	; Returns - Cb = 0, OK
	;           Cb = 1, error

PROCREC	JSR CHRIN
	CMP #$20
	BCC PROCREC		; Skip CR/LF
	CMP #':'		; Check start of record
	BNE RECINV
	JSR GETHEXB		; Read record length
	STA CSUM
	CMP #$11
	BCS REC2LNG
	STA RECLEN
	JSR GETHEXW		; Read address
	STA ADDR
	STY ADDR+1
	CLC
	ADC CSUM
	CLC
	ADC ADDR+1
	STA CSUM
	LDA START		; Preserve first address
	ORA START+1
	BNE GETTYP
	LDA ADDR
	STA START
	LDA ADDR+1
	STA START+1
GETTYP	JSR GETHEXB		; Read record type
	PHA
	CLC
	ADC CSUM
	STA CSUM
	PLA
	BNE ROVER1
	JMP DATAREC
ROVER1	CMP #$01
	BNE ROVER2
	JMP ENDREC
ROVER2	CMP #0
	BPL ROVER3
	JMP EXTREC
ROVER3	LDA #<RTINVS
	LDY #>RTINVS
	JSR PRTSTR
	SEC
	RTS

RECINV	LDA #<RINVS
	LDY #>RINVS
	JSR PRTSTR
	SEC
	RTS
REC2LNG	LDA #<R2LNGS
	LDY #>R2LNGS
	JSR PRTSTR
	SEC
	RTS
CSUMBAD	LDA #<CSBADS
	LDY #>CSBADS
	JSR PRTSTR
	SEC
	RTS


	; Process Data record

DATAREC	LDY #0
DRLOOP	STY POS
	JSR GETHEXB		; Read data
	PHA
	CLC
	ADC CSUM
	STA CSUM
	PLA
	LDY POS
	STA (ADDR),Y
	INY
	CPY RECLEN
	BCC DRLOOP
	JSR GETHEXB		; Read checksum
	CLC
	ADC CSUM
	BNE CSUMBAD		; Verify checksum
	LDA #<ROKS
	LDY #>ROKS
	JSR PRTSTR
	CLC
	LDA RECLEN
	ADC ADDR
	STA END
	LDA ADDR+1
	ADC #0
	STA END+1
	CLC
	RTS


	; Process End of File record

ENDREC	JSR GETHEXB		; Read checksum
	CLC
	ADC CSUM
	BNE CSUMBAD		; Verify checksum
	LDA #<EOTS
	LDY #>EOTS
	JSR PRTSTR
	SEC
	RTS


	; Non-standard records

FNAM	= $0200

EXTREC	CMP #$81
	BNE RECINV
	LDY #0
SVLOOP	STY POS
	JSR GETHEXB		; Read data
	PHA
	CLC
	ADC CSUM
	STA CSUM
	PLA
	LDY POS
	STA FNAM,Y
	INY
	CPY RECLEN
	BCC SVLOOP
	TYA
	LDX #<FNAM
	LDY #>FNAM
	JSR SETNAM
	JSR GETHEXB		; Read checksum
	LDA #1
	LDX ADDR
	LDY #$FF
	JSR SETLFS
	LDA #START
	LDX END
	LDY END+1
	JSR SAVE
	PHP
	PHA
	LDA #13
	JSR CHROUT
	PLA
	PLP
	BCS SVERR
	RTS

SVERR	LDA #<SVERRS
	LDY #>SVERRS
	JSR PRTSTR
	SEC
	RTS

	
	; Read a 8-bit hex value.
	;
	; Returns - .A - value

HXVAL	= $4E

GETHEXB	LDY #$00
	STY HXVAL
GBLOOP	JSR CHRIN
	CMP #"0"
	BCC GBCVRT
	CMP #"9"+1
	BCS GBXDIG
	SEC
	SBC #"0"
GBADD	ASL HXVAL
	ASL HXVAL
	ASL HXVAL
	ASL HXVAL
	ORA HXVAL
	STA HXVAL
	INY
	CPY #$02
	BCC GBLOOP
GBCVRT	LDA HXVAL
	RTS
GBXDIG	CMP #"A"
	BCC GBCVRT
	CMP #"F"+1
	BCS GBCVRT
	SEC
	SBC #"A"-10
	BCS GBADD


	; Read a 16-bit hex value.
	;
	; Returns - .A - LSB
	; 	    .Y - MSB

GETHEXW	LDY #$00
	STY HXVAL
	STY HXVAL+1
GWLOOP	JSR CHRIN
	CMP #"0"
	BCC GWCVRT
	CMP #"9"+1
	BCS GWXDIG
	SEC
	SBC #"0"
GWADD	ASL HXVAL
	ROL HXVAL+1
	ASL HXVAL
	ROL HXVAL+1
	ASL HXVAL
	ROL HXVAL+1
	ASL HXVAL
	ROL HXVAL+1
	ORA HXVAL
	STA HXVAL
	INY
	CPY #$04
	BCC GWLOOP
GWCVRT	LDA HXVAL
	LDY HXVAL+1
	RTS
GWXDIG	CMP #"A"
	BCC GWCVRT
	CMP #"F"+1
	BCS GWCVRT
	SEC
	SBC #"A"-10
	BCS GWADD


	; Print a NUL terminated string.
	;
	; .A - LSB
	; .Y - MSB

STRPTR	= $4D

PRTSTR	STA STRPTR
	STY STRPTR+1
	LDY #$00
	LDA (STRPTR),Y
PSLOOP	JSR CHROUT
	INY
	LDA (STRPTR),Y
	BNE PSLOOP
	RTS
	
	; Serial port parameters
SERCTRL	.BYTE %00001000		; 1 stop bit, 8 data bits, 1200 baud
SERCMD	.BYTE %00000000		; no parity, full duplex, 3 line

OPFS	.ASC "SERIAL OPEN FAILED.", 13, 0
CKFS	.ASC "CHANNEL OPEN FAILED.", 13, 0
RINVS	.ASC "RECORD INVALID.", 13, 0
R2LNGS	.ASC "RECORD TOO LONG.", 13, 0
RTINVS	.ASC "RECORD TYPE INVALID.", 13, 0
ROKS	.ASC "RECORD OK", 13, 0
CSBADS	.ASC "CHECKSUM MISMATCH.", 13, 0
EOTS	.ASC "END OF TRANSFER.", 13, 0
SVERRS	.ASC "SAVE ERROR.", 13, 0
